
from contextlib import contextmanager
from unittest.mock import patch

from odoo import api
from odoo.tests import tagged

from odoo.addons.loempia.tests.common import MockRequest, TestLoempia
from odoo.addons.loempia.controllers.frontend import LoempiaDisputesController

from datetime import date
from dateutil.relativedelta import relativedelta

@tagged('post_install', '-at_install')
class TestLoempiaRefund(TestLoempia):

    @classmethod
    def setUpClass(self):
        super(TestLoempiaRefund, self).setUpClass()
        self.Controller = LoempiaDisputesController()
        self.registry.enter_test_mode(self.cr)

    @classmethod
    def tearDownClass(self):
        self.registry.leave_test_mode()
        super(TestLoempiaRefund, self).tearDownClass()

    @contextmanager
    def mockAccountGateway(self):
        self._sms = []

        def _create_support_ticket(local_endpoint, params):
            # mock ticket creation during refund validation
            return True

        try:
            with patch.object(LoempiaDisputesController, '_create_support_ticket', side_effect=_create_support_ticket):
                yield
        finally:
            pass

    def _create_refund(self, request_limit_date, purchase):
        refund = self.env['loempia.refund'].create({
            'purchase_id': purchase.id,
            'instigator_id': self.functional.id,
            'request_limit_date': request_limit_date,
            'state': 'pending',
            'reason': 'other'
        })
        purchase.refund_id = refund.id
        return refund

    def test_01_refund_request(self):
        """
        Test to create a refund, with a user which is allowed to perform this action.
        The two first test are supposed to fail as there are missing information in
        order to create the refund request. The third test should ensure that a refund
        request can't be opened after 60 days. The fourth test contains all the information
        and should succeed in the creation of a refund request.
        """

        env = api.Environment(self.env.cr, self.maintainer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            # 01
            self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, '', reason=1, other='')
            self.assertEqual(len(self.purchase_A.refund_id), 0, msg="A commment is necessary to open a refund request.")

            # 02
            self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, '', reason=None, other='')
            self.assertEqual(len(self.purchase_A.refund_id), 0, msg="A reason is necessary to open a refund request.")

            # 03 change the po date to fail creating the refund request
            self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
            nb_days = int(self.env['ir.config_parameter'].sudo().get_param('loempia.max_claim_delay', '60')) + 1
            self.purchase_A.sudo().date_order = date.today() - relativedelta(days=nb_days)
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_01_refund_request_maintainer', reason=1, other='')
            self.assertEqual(len(self.purchase_A.refund_id), 0, msg="A refund request cannot be opened after the max claim delays.")
            self.purchase_A.sudo().date_order = date.today()

            # 04
            self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_01_refund_request_maintainer', reason=1, other='')
            self.assertEqual(
                len(self.purchase_A.refund_id),
                1,
                msg="A maintainer needs to be able to create a refund associated to a purchase of one of his/her app."
            )
            self.assertEqual(
                self.purchase_A.refund_id.maintainer_message,
                'test_01_refund_request_maintainer',
                msg="The maintainer's message should be identical to the posted comment."
            )
            self.assertEqual(
                self.purchase_A.refund_id.instigator_id.id,
                self.maintainer.id,
                msg="The maintainer's repository must be able to request a refund."
            )

    def test_02_refund_request(self):
        """
        This test should ensure that a functional agent is able to open a refund
        request.
        """

        env = api.Environment(self.env.cr, self.functional.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            self.assertEqual(r.env.user.id, self.functional.id, msg='Ensure correct user')
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_02_refund_request_functional', reason=1, other='')
            self.assertEqual(len(self.purchase_A.refund_id), 1, msg="A refund must be associated to the created purchase")
            self.assertEqual(
                self.purchase_A.refund_id.maintainer_message,
                'test_02_refund_request_functional',
                msg="The maintainer message should be identical to the posted comment"
            )
            self.assertEqual(
                self.purchase_A.refund_id.instigator_id.id,
                self.functional.id,
                msg="A functional agent must be able to request a refund."
            )

    def test_03_refund_request(self):
        """
        This test should ensure that a customer is not able to open a refund request.
        """
        env = api.Environment(self.env.cr, self.customer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            self.assertEqual(r.env.user.id, self.customer.id, msg='Ensure correct user')
            self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_03_refund_request_portal', reason=1, other='')
            self.assertEqual(len(self.purchase_A.refund_id), 0, msg="Customer should not be able to request a refund.")

    def test_01_extend_refund(self):
        """
        The first test consists to ensure that a functional agent can extend the limit date of
        an expired refund. However, he cannot extend it if the refund is not expired (second test).
        """
        request_limit_date = date.today() - relativedelta(days=2)
        refund = self._create_refund(request_limit_date, self.purchase_A)

        env = api.Environment(self.env.cr, self.functional.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            # 01
            self.assertEqual(r.env.user.id, self.functional.id, msg='Ensure correct user')
            self.Controller.extend_refund(refund.id)
            limit_date = date.today() + relativedelta(days=5)
            self.assertFalse(refund.expired, msg="After extension, the refund should not be expired anymore.")
            self.assertEqual(
                refund.request_limit_date, limit_date,
                msg="If a functional agent extends a valid refund request, the request limit date has to be greater than today."
            )

            # 02
            self.assertEqual(r.env.user.id, self.functional.id, msg='Ensure correct user')
            self.Controller.extend_refund(refund.id)
            self.assertFalse(
                refund.expired,
                msg="As the refund was not expired before the extension, it should remains the same after it."
            )
            self.assertEqual(
                refund.request_limit_date, limit_date,
                msg="Extending an non-expired refund request should not change the request_limit_date field"
            )

    def test_02_extend_refund(self):
        """
        This test ensures that a portal user is not allowed to extend the request limit date of
        a refund request.
        """
        request_limit_date = date.today() - relativedelta(days=2)
        refund = self._create_refund(request_limit_date, self.purchase_A)

        env = api.Environment(self.env.cr, self.customer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            self.assertEqual(r.env.user.id, self.customer.id, msg='Ensure correct user')
            self.Controller.extend_refund(refund.id)
            self.assertTrue(refund.expired, msg="After an invalid extension, the refund should remain expired")
            self.assertEqual(
                refund.request_limit_date, request_limit_date,
                msg="After an invalid extension, the request_limit_date should remain the same"
            )

    def test_03_extend_refund(self):
        """
        This test ensures that an app's maintainer is not allowed to extend the request limit date of
        a refund request.
        """
        request_limit_date = date.today() - relativedelta(days=2)
        refund = self._create_refund(request_limit_date, self.purchase_A)

        env = api.Environment(self.env.cr, self.maintainer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
            self.Controller.extend_refund(refund.id)
            self.assertTrue(refund.expired, msg="After an invalid extension, the refund should remain expired")
            self.assertEqual(
                refund.request_limit_date, request_limit_date,
                msg="After an invalid extension, the request_limit_date should remain the same"
            )

    def test_01_validate_refund(self):
        """
        The first test ensures that an expired refund cannot be validated.
        The second test ensures that the customer can validate its refund.
        """
        refund = self._create_refund(date.today() - relativedelta(days=31), self.purchase_A)

        env = api.Environment(self.env.cr, self.customer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            with self.mockAccountGateway():
                # 01
                self.assertEqual(r.env.user.id, self.customer.id, msg='Ensure correct user')
                self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_01_validate_refund', reason=None, other='')
                self.assertEqual(
                    refund.state,
                    'pending',
                    msg="It's impossible to validate an expired refund."
                )
                refund.request_limit_date = date.today() + relativedelta(days=31)

            # with self.mockAccountGateway():
                #02
                self.assertEqual(r.env.user.id, self.customer.id, msg='Ensure correct user')
                self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_01_validate_refund', reason=None, other='')
                self.assertEqual(
                    refund.user_message,
                    'test_01_validate_refund',
                    msg="Message of the user is correctly set on the purchase."
                )
                self.assertEqual(
                    refund.state,
                    'accepted',
                    msg="Validate a refund should mark it as accepted"
                )

    def test_02_validate_refund(self):
        """
        This test ensures that a functional agent can validate a refund.
        """
        refund = self._create_refund(date.today() + relativedelta(days=31), self.purchase_A)

        env = api.Environment(self.env.cr, self.functional.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            with self.mockAccountGateway():
                self.assertEqual(r.env.user.id, self.functional.id, msg='Ensure correct user')
                self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_02_validate_refund', reason=None, other='')
                self.assertEqual(
                    refund.user_message,
                    'test_02_validate_refund',
                    msg="Message of the user is correctly set on the purchase."
                )
                self.assertEqual(
                    refund.state,
                    'accepted',
                    msg="Validate a refund should mark it as accepted."
                )

    def test_03_validate_refund(self):
        """
        This test ensures that app's maintainer cannot validate a refund.
        """
        refund = self._create_refund(date.today() + relativedelta(days=31), self.purchase_A)

        env = api.Environment(self.env.cr, self.maintainer.id, {})
        with MockRequest(env, website=self.website.with_env(env)) as r:
            with self.mockAccountGateway():
                self.assertEqual(r.env.user.id, self.maintainer.id, msg='Ensure correct user')
                self.Controller.refund(self.purchase_A.order_id, self.purchase_A.module_id.id, 'test_03_validate_refund', reason=None, other='')
                self.assertEqual(
                    refund.state,
                    'pending',
                    msg="A maintainer cannot validate an open refund request."
                )
